# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from tempest.lib import decorators

from trove_tempest_plugin.tests import constants
from trove_tempest_plugin.tests.scenario import base_backup
from trove_tempest_plugin.tests import utils


class TestBackupMySQL(base_backup.TestBackupBase):
    datastore = 'mysql'

    @classmethod
    def insert_data(cls, ip, username=constants.DB_USER,
                    password=constants.DB_PASS, database=constants.DB_NAME,
                    **kwargs):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "CREATE TABLE Persons (ID int, String varchar(255));",
                "insert into Persons VALUES (1, 'Lingxian Kong');",
            ]
            db_client.mysql_execute(cmds)

    @classmethod
    def insert_data_inc(cls, ip, username=constants.DB_USER,
                        password=constants.DB_PASS, database=constants.DB_NAME,
                        **kwargs):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "insert into Persons VALUES (99, 'OpenStack');"
            ]
            db_client.mysql_execute(cmds)

    def verify_data(self, ip, username=constants.DB_USER,
                    password=constants.DB_PASS, database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from Persons;"
            ret = db_client.mysql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertEqual(1, len(rows))

        result = dict(zip(keys, rows[0]))
        expected = {'ID': 1, 'String': 'Lingxian Kong'}
        self.assertEqual(expected, result)

    def verify_data_inc(self, ip, username=constants.DB_USER,
                        password=constants.DB_PASS,
                        database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from Persons;"
            ret = db_client.mysql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertEqual(2, len(rows))

        actual = []
        for index in range(2):
            actual.append(dict(zip(keys, rows[index])))

        expected = [
            {'ID': 1, 'String': 'Lingxian Kong'},
            {'ID': 99, 'String': 'OpenStack'},
        ]
        self.assertEqual(expected, actual)

    @decorators.idempotent_id("b90626ae-f412-11ea-a950-00224d6b7bc1")
    def test_backup_full(self):
        self.backup_full_test()

    @decorators.idempotent_id("f8f985c2-ae02-11ea-b87c-00224d6b7bc1")
    def test_backup_incremental(self):
        self.backup_incremental_test()


class TestBackupPostgreSQL(base_backup.TestBackupBase):
    datastore = 'postgresql'
    create_user = False
    enable_root = True
    root_password = ""

    @classmethod
    def insert_data(cls, ip):
        db_url = (f'postgresql+psycopg2://root:{cls.password}@'
                  f'{ip}:5432/postgres')
        with utils.SQLClient(db_url) as db_client:
            cmd = "CREATE DATABASE testdb;"
            db_client.pgsql_execute(cmd)

        db_url = (f'postgresql+psycopg2://root:{cls.password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "CREATE TABLE persons (id INT PRIMARY KEY NOT NULL, "
                "string VARCHAR(255));",
                "INSERT INTO persons (id,string) VALUES (1, 'Lingxian Kong');",
            ]
            db_client.pgsql_execute(cmds)

    @classmethod
    def insert_data_inc(cls, ip):
        db_url = (f'postgresql+psycopg2://root:{cls.password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "INSERT INTO persons (id,string) VALUES (99, 'OpenStack');"
            ]
            db_client.pgsql_execute(cmds)

    def verify_data(self, ip):
        db_url = (f'postgresql+psycopg2://root:{self.root_password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from persons;"
            ret = db_client.pgsql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertEqual(1, len(rows))

        result = dict(zip(keys, rows[0]))
        expected = {'id': 1, 'string': 'Lingxian Kong'}
        self.assertEqual(expected, result)

    def verify_data_inc(self, ip, username=constants.DB_USER,
                        password=constants.DB_PASS,
                        database=constants.DB_NAME):
        db_url = (f'postgresql+psycopg2://root:{self.root_password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from persons;"
            ret = db_client.pgsql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertEqual(2, len(rows))

        actual = []
        for index in range(2):
            actual.append(dict(zip(keys, rows[index])))

        expected = [
            {'id': 1, 'string': 'Lingxian Kong'},
            {'id': 99, 'string': 'OpenStack'},
        ]
        self.assertEqual(expected, actual)

    @decorators.idempotent_id("e8339fce-f412-11ea-a950-00224d6b7bc1")
    def test_backup_full(self):
        self.backup_full_test()

    @decorators.idempotent_id("ec387400-f412-11ea-a950-00224d6b7bc1")
    def test_backup_incremental(self):
        self.backup_incremental_test()



# Copyright (c) 2016 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

service_option = cfg.BoolOpt(
    'trove',
    default=True,
    help="Whether or not Trove is expected to be available"
)

database_group = cfg.OptGroup(
    name='database',
    title='Database Service Options'
)

DatabaseGroup = [
    cfg.StrOpt(
        'catalog_type',
        default='database',
        help="Catalog type of the Database service."
    ),
    cfg.StrOpt(
        'endpoint_type',
        default='publicURL',
        choices=['public', 'admin', 'internal', 'publicURL', 'adminURL',
                 'internalURL'],
        help="The endpoint type to use for the Database service."
    ),
    cfg.ListOpt(
        'enabled_datastores',
        default=['mysql']
    ),
    cfg.DictOpt(
        'default_datastore_versions',
        default={'mysql': '5.7.29'},
        help='The default datastore versions used to create instance',
    ),
    cfg.DictOpt(
        'pre_upgrade_datastore_versions',
        default={},
        help='The datastore versions used to create instances that need to be '
             'upgrade.',
    ),
    cfg.IntOpt('database_build_timeout',
               default=1800,
               help='Timeout in seconds to wait for a database instance to '
                    'build.'),
    cfg.IntOpt(
        'database_restore_timeout',
        default=3600,
        help='Timeout in seconds to wait for a database instance to '
             'be restored.'
    ),
    cfg.IntOpt(
        'backup_wait_timeout',
        default=600,
        help='Timeout in seconds to wait for a backup to be completed.'
    ),
    cfg.StrOpt(
        'flavor_id',
        default="d2",
        help="The Nova flavor ID used for creating database instance."
    ),
    cfg.StrOpt(
        'resize_flavor_id',
        default="d3",
        help="The Nova flavor ID used for resizing database instance."
    ),
    cfg.StrOpt(
        'shared_network',
        default="private",
        help=('Pre-defined network name or ID used for creating database '
              'instance.')
    ),
    cfg.StrOpt(
        'subnet_cidr',
        default='10.1.1.0/24',
        help=('The Neutron CIDR format subnet to use for database network '
              'creation.')
    ),
    cfg.StrOpt(
        'volume_type',
        default="lvmdriver-1",
        help="The Cinder volume type used for creating database instance."
    ),
    cfg.StrOpt(
        'database_log_container',
        default="database_logs",
        help="The name of Swift container for the database instance log, "
             "should be the same with the config in the cloud."
    ),
    cfg.BoolOpt(
        'remove_swift_account',
        default=True,
        help='If clean up the Swift account for test users after backup '
             'function test. ResellerAdmin should be added to tempest config '
             'option "tempest_roles".'
    ),
    cfg.StrOpt(
        'rebuild_image_id',
        help="The new guest image ID used for rebuilding database instance."
    ),
]



# Copyright (c) 2016 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from tempest import config
from tempest.test_discover import plugins

from trove_tempest_plugin import config as trove_config


class TroveTempestPlugin(plugins.TempestPlugin):

    def load_tests(self):
        base_path = os.path.split(os.path.dirname(
            os.path.abspath(__file__)))[0]
        test_dir = "trove_tempest_plugin/tests"
        full_test_dir = os.path.join(base_path, test_dir)
        return full_test_dir, base_path

    def register_opts(self, conf):
        conf.register_group(trove_config.database_group)
        conf.register_opts(trove_config.DatabaseGroup, group='database')
        conf.register_opt(trove_config.service_option,
                          group='service_available')

    def get_opt_lists(self):
        return [('database', trove_config.DatabaseGroup),
                ('service_available', [trove_config.service_option])]

    def get_service_clients(self):
        service_config = config.service_client_config('database')
        service_params = {
            'name': 'database',
            'service_version': 'database',
            'module_path': 'trove_tempest_plugin.services.client',
            'client_names': ['TroveClient']
        }
        service_params.update(service_config)
        return [service_params]



# Copyright 2018 Samsung Electronics
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from urllib import parse as urlparse

from oslo_serialization import jsonutils as json
import tenacity

from tempest.lib.common import rest_client
from tempest.lib import exceptions


class TroveClient(rest_client.RestClient):
    def __init__(self, auth_provider, **kwargs):
        super(TroveClient, self).__init__(auth_provider, **kwargs)

    def get_resource(self, obj, id, expected_status_code=200):
        url = '/%s/%s' % (obj, id)
        resp, body = self.get(url)
        self.expected_success(expected_status_code, resp.status)

        return rest_client.ResponseBody(resp, json.loads(body))

    def list_resources(self, obj, expected_status_code=200, **filters):
        url = '/%s' % obj
        if filters:
            # Encode provided dict of fields into a series of key=value pairs
            # separated by '&' characters.
            #
            # The field value can be a sequence. Setting option doseq to True
            # enforces producing individual key-value pair for each element of
            # the sequence under the same key.
            #
            # e.g. {'foo': 'bar', 'baz': ['test1', 'test2']}
            #      => foo=bar&baz=test1&baz=test2
            url += '?' + urlparse.urlencode(filters, doseq=True)

        resp, body = self.get(url)
        self.expected_success(expected_status_code, resp.status)

        return rest_client.ResponseBody(resp, json.loads(body))

    @tenacity.retry(
        wait=tenacity.wait_fixed(5),
        stop=tenacity.stop_after_attempt(2),
        retry=tenacity.retry_if_exception_type(),
        reraise=True
    )
    def delete_resource(self, obj, id, ignore_notfound=False):
        try:
            resp, _ = self.delete('/{obj}/{id}'.format(obj=obj, id=id))
            self.expected_success(202, resp.status)
            return resp
        except exceptions.NotFound:
            if ignore_notfound:
                return None
            else:
                raise

    def force_delete_instance(self, id):
        headers = {"Content-Type": "application/json"}
        body = {"reset_status": {}}
        resp, _ = self.post(f'/instances/{id}/action', json.dumps(body),
                            headers=headers)
        self.expected_success(202, resp.status)

        self.delete_resource('instances', id, ignore_notfound=True)

    def create_resource(self, obj, req_body, extra_headers={},
                        expected_status_code=200, need_response=True):
        headers = {"Content-Type": "application/json"}
        headers = dict(headers, **extra_headers)
        url = '/%s' % obj

        resp, body = self.post(url, json.dumps(req_body), headers=headers)
        self.expected_success(expected_status_code, resp.status)

        if need_response:
            return rest_client.ResponseBody(resp, json.loads(body))

    def patch_resource(self, obj, id, req_body, expected_status_code=202):
        url = '/{obj}/{id}'.format(obj=obj, id=id)
        headers = {"Content-Type": "application/json"}

        resp, _ = self.patch(url, json.dumps(req_body), headers=headers)
        self.expected_success(expected_status_code, resp.status)

    def put_resource(self, url, req_body, expected_status_code=202):
        url = '/%s' % url
        headers = {"Content-Type": "application/json"}

        resp, _ = self.put(url, json.dumps(req_body), headers=headers)
        self.expected_success(expected_status_code, resp.status)



# Copyright 2019 Catalyst Cloud Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from oslo_log import log as logging
from tempest import config
from tempest.lib import decorators
from tempest.lib import exceptions

from trove_tempest_plugin.tests import base as trove_base
from trove_tempest_plugin.tests import constants
from trove_tempest_plugin.tests import utils

CONF = config.CONF
LOG = logging.getLogger(__name__)


class TestInstanceBasicBase(trove_base.BaseTroveTest):
    def get_config_value(self, ip, option, **kwargs):
        pass

    def configuration_test(self, create_values, update_values,
                           need_restart=False):
        """Test configuration.

        The create_values and update_values are both dict with one key, the
        value should be in type int.
        """
        # Create new configuration
        config_name = self.get_resource_name('config')
        key = list(create_values.keys())[0]
        value = list(create_values.values())[0]

        LOG.info(f"Creating new configuration {config_name}")
        config = self.create_config(
            config_name, create_values, self.datastore,
            self.instance['datastore']['version'])
        config_id = config['configuration']['id']
        self.addCleanup(self.client.delete_resource, 'configurations',
                        config_id, ignore_notfound=True)
        self.assertEqual(0, config['configuration']['instance_count'])

        ret = self.client.list_resources(
            f"configurations/{config_id}/instances")
        self.assertEqual(0, len(ret['instances']))

        # Attach the configuration to the existing instance
        LOG.info(f"Attaching config {config_id} to instance "
                 f"{self.instance_id}")
        self.attach_config(self.instance_id, config_id)

        if need_restart:
            LOG.info(f"Restarting instance {self.instance_id}")
            self.restart_instance(self.instance_id)

        ret = self.client.list_resources(
            f"configurations/{config_id}/instances")
        self.assertEqual(1, len(ret['instances']))
        self.assertEqual(self.instance_id, ret['instances'][0]['id'])

        # Get new config option value
        LOG.info(f"Getting config value for {key} on {self.instance_ip}")
        cur_value = self.get_config_value(self.instance_ip, key)
        self.assertEqual(value, cur_value)

        # Update configuration
        new_key = list(update_values.keys())[0]
        new_value = list(update_values.values())[0]
        patch_config = {
            "configuration": {
                "values": update_values
            }
        }
        LOG.info(f"Updating config {config_id}")
        self.client.patch_resource('configurations', config_id, patch_config,
                                   expected_status_code=200)

        if need_restart:
            LOG.info(f"Restarting instance {self.instance_id}")
            self.restart_instance(self.instance_id)

        LOG.info(f"Getting config value for {new_key} on {self.instance_ip}")
        cur_value = self.get_config_value(self.instance_ip, new_key)
        self.assertEqual(new_value, cur_value)

        # Detach the configuration from the instance
        LOG.info(f"Detaching from instance {self.instance_id}")
        self.detach_config(self.instance_id)

        if need_restart:
            LOG.info(f"Restarting instance {self.instance_id}")
            self.restart_instance(self.instance_id)

        ret = self.client.list_resources(
            f"configurations/{config_id}/instances")
        self.assertEqual(0, len(ret['instances']))

        # Get new config option value
        LOG.info(f"Getting config value for {new_key} on {self.instance_ip}")
        cur_value = self.get_config_value(self.instance_ip, new_key)
        self.assertNotEqual(value, cur_value)
        self.assertNotEqual(new_value, cur_value)


class TestInstanceBasicMySQLBase(TestInstanceBasicBase):
    def _access_db(self, ip, username=constants.DB_USER,
                   password=constants.DB_PASS, database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmd = "SELECT 1;"
            db_client.mysql_execute(cmd)

    def get_config_value(self, ip, option, username=constants.DB_USER,
                         password=constants.DB_PASS):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306'
        with utils.SQLClient(db_url) as db_client:
            cmd = f"show variables where Variable_name in ('{option}');"
            ret = db_client.mysql_execute(cmd)
            rows = ret.fetchall()
        self.assertEqual(1, len(rows))
        return int(rows[0][1])

    @decorators.idempotent_id("40cf38ce-cfbf-11e9-8760-1458d058cfb2")
    def test_database_access(self):
        databases = self.get_databases(self.instance_id)
        db_names = [db['name'] for db in databases]
        self.assertIn(constants.DB_NAME, db_names)

        users = self.get_users(self.instance_id)
        user_names = [user['name'] for user in users]
        self.assertIn(constants.DB_USER, user_names)

        LOG.info(f"Accessing database on {self.instance_ip}")
        self._access_db(self.instance_ip)

    @decorators.idempotent_id("c5a9dcda-af5b-11ea-b87c-00224d6b7bc1")
    def test_user_database(self):
        db1 = 'foo'
        db2 = 'bar'
        user1 = 'foo_user'
        user2 = 'bar_user'

        users = self.get_users(self.instance_id)
        cur_user_names = [user['name'] for user in users]
        self.assertNotIn(user1, cur_user_names)
        self.assertNotIn(user2, cur_user_names)

        databases = self.get_databases(self.instance_id)
        cur_db_names = [db['name'] for db in databases]
        self.assertNotIn(db1, cur_db_names)
        self.assertNotIn(db2, cur_db_names)

        LOG.info(f"Creating databases in instance {self.instance_id}")
        create_db = {
            "databases": [
                {
                    "name": db1
                },
                {
                    "name": db2
                }
            ]
        }
        self.client.create_resource(f"instances/{self.instance_id}/databases",
                                    create_db, expected_status_code=202,
                                    need_response=False)

        databases = self.get_databases(self.instance_id)
        cur_db_names = [db['name'] for db in databases]
        self.assertIn(db1, cur_db_names)
        self.assertIn(db2, cur_db_names)

        LOG.info(f"Creating users in instance {self.instance_id}")
        create_user = {
            "users": [
                {
                    "databases": [
                        {
                            "name": db1
                        }
                    ],
                    "name": user1,
                    "password": constants.DB_PASS
                },
                {
                    "name": user2,
                    "password": constants.DB_PASS
                }
            ]
        }
        self.client.create_resource(f"instances/{self.instance_id}/users",
                                    create_user, expected_status_code=202,
                                    need_response=False)

        users = self.get_users(self.instance_id)
        cur_user_names = [user['name'] for user in users]
        self.assertIn(user1, cur_user_names)
        self.assertIn(user2, cur_user_names)

        # user1 should have access to db1
        LOG.info(f"Accessing database on {self.instance_ip}, user: {user1}, "
                 f"db: {db1}")
        self._access_db(self.instance_ip, user1, constants.DB_PASS, db1)
        # user2 should not have access to db2
        self.assertRaises(exceptions.TempestException, self._access_db,
                          self.instance_ip, user2, constants.DB_PASS, db2)

        LOG.info(f"Granting user {user2} access to database {db2}")
        grant_access = {
            "databases": [
                {
                    "name": db2
                }
            ]
        }
        self.client.put_resource(
            f'/instances/{self.instance_id}/users/{user2}/databases',
            grant_access)
        user2_dbs = self.client.list_resources(
            f'instances/{self.instance_id}/users/{user2}/databases')
        user2_dbs = [db['name'] for db in user2_dbs['databases']]
        self.assertIn(db2, user2_dbs)
        # Now user2 should have access to db2
        LOG.info(f"Accessing database on {self.instance_ip}, user: {user2}, "
                 f"db: {db2}")
        self._access_db(self.instance_ip, user2, constants.DB_PASS, db2)

        LOG.info(f"Revoking user {user2} access to database {db2}")
        self.client.delete_resource(
            f'instances/{self.instance_id}/users/{user2}/databases', db2)
        # user2 should not have access to db2
        self.assertRaises(exceptions.TempestException, self._access_db,
                          self.instance_ip, user2, constants.DB_PASS, db2)

        LOG.info(f"Deleting user {user2}")
        self.client.delete_resource(
            f'instances/{self.instance_id}/users', user2)
        users = self.get_users(self.instance_id)
        cur_user_names = [user['name'] for user in users]
        self.assertIn(user1, cur_user_names)
        self.assertNotIn(user2, cur_user_names)

        LOG.info(f"Deleting database {db2}")
        self.client.delete_resource(
            f'instances/{self.instance_id}/databases', db2)
        databases = self.get_databases(self.instance_id)
        cur_db_names = [db['name'] for db in databases]
        self.assertIn(db1, cur_db_names)
        self.assertNotIn(db2, cur_db_names)

    @decorators.idempotent_id("ce8277b0-af7c-11ea-b87c-00224d6b7bc1")
    def test_configuration(self):
        create_values = {"max_connections": 555}
        update_values = {"max_connections": 666}
        self.configuration_test(create_values, update_values)
